@@ -28,6 +28,8 @@ COMP_WORDBREAKS = " \t\n\"'><=;|&(:"
 
 BASH_VERSION = subprocess.check_output(['bash', '-c', 'echo $BASH_VERSION']).decode()
 BASH_MAJOR_VERSION = int(BASH_VERSION.split('.')[0])
+FISH_VERSION_STR = subprocess.check_output(['fish', '-c', 'echo -n $FISH_VERSION']).decode()
+FISH_VERSION_TUPLE = tuple(int(x) for x in FISH_VERSION_STR.split('.'))
 
 
 class TempDir(object):
@@ -74,8 +76,10 @@ class TestArgcomplete(unittest.TestCase):
         with TemporaryFile() as t:
             os.environ["COMP_LINE"] = ensure_bytes(command) if USING_PYTHON2 else command
             os.environ["COMP_POINT"] = point
-            self.assertRaises(SystemExit, completer, parser, output_stream=t,
-                              exit_method=sys.exit, **kwargs)
+            with self.assertRaises(SystemExit) as cm:
+                completer(parser, output_stream=t, exit_method=sys.exit, **kwargs)
+            if cm.exception.code != 0:
+                raise Exception("Unexpected exit code %d" % cm.exception.code)
             t.seek(0)
             return t.read().decode(sys_encoding).split(IFS)
 
@@ -371,7 +375,7 @@ class TestArgcomplete(unittest.TestCase):
              [r"on\ a\ train", r"with\ a\ goat", r"on\ a\ boat", r"in\ the\ rain", "--help", "-h"]),
             ("prog eggs ", [r"on\ a\ train", r"with\ a\ goat", r"on\ a\ boat", r"in\ the\ rain", "--help", "-h"]),
             ("prog eggs \"on a", ['on a train', 'on a boat']),
-            ("prog eggs on\\ a", [r"on\ a\ train", "on\ a\ boat"]),
+            ("prog eggs on\\ a", [r"on\ a\ train", r"on\ a\ boat"]),
             ("prog spam ", ["iberico", "ham", "--help", "-h"]),
         )
 
@@ -398,7 +402,7 @@ class TestArgcomplete(unittest.TestCase):
             ("prog ", ["--книга", "-h", "--help"]),
             ("prog --книга ",
              [r"Трудно\ быть\ богом", r"Парень\ из\ преисподней", r"Понедельник\ начинается\ в\ субботу"]),
-            ("prog --книга П", [r"Парень\ из\ преисподней", "Понедельник\ начинается\ в\ субботу"]),
+            ("prog --книга П", [r"Парень\ из\ преисподней", r"Понедельник\ начинается\ в\ субботу"]),
             ("prog --книга Пу", [""]),
         )
 
@@ -504,6 +508,41 @@ class TestArgcomplete(unittest.TestCase):
         self.assertEqual("ttt", disp.get("--oh", ""))
         self.assertEqual("ccc", disp.get("--ch", ""))
 
+    @unittest.skipIf(USING_PYTHON2, "Subparser aliases aren't supported on Python 2")
+    def test_display_completions_with_aliases(self):
+        parser = ArgumentParser()
+        parser.add_subparsers().add_parser("a", aliases=["b", "c"], help="abc help")
+
+        # empty
+        completer = CompletionFinder(parser)
+        completer.rl_complete("", 0)
+        disp = completer.get_display_completions()
+        self.assertEqual({"a b c": "abc help", "-h --help": "show this help message and exit"}, disp)
+
+        # a
+        completer = CompletionFinder(parser)
+        completer.rl_complete("a", 0)
+        disp = completer.get_display_completions()
+        self.assertEqual({"a": "abc help", "": "show this help message and exit"}, disp)
+
+        # b
+        completer = CompletionFinder(parser)
+        completer.rl_complete("b", 0)
+        disp = completer.get_display_completions()
+        self.assertEqual({"b": "abc help", "": "show this help message and exit"}, disp)
+
+        # c
+        completer = CompletionFinder(parser)
+        completer.rl_complete("c", 0)
+        disp = completer.get_display_completions()
+        self.assertEqual({"c": "abc help", "": "show this help message and exit"}, disp)
+
+        # (
+        completer = CompletionFinder(parser)
+        completer.rl_complete("(", 0)
+        disp = completer.get_display_completions()
+        self.assertEqual({"": "show this help message and exit"}, disp)
+
     def test_nargs_one_or_more(self):
         def make_parser():
             parser = ArgumentParser()
@@ -753,9 +792,41 @@ class TestArgcomplete(unittest.TestCase):
             fh.write(sc.encode())
             fh.flush()
             subprocess.check_call(['bash', '-n', fh.name])
+        with NamedTemporaryFile() as fh:
+            sc = shellcode(["prog"], use_defaults=True, shell="bash", complete_arguments=None,
+                           argcomplete_script="~/.bash_completion.d/prog.py")
+            fh.write(sc.encode())
+            fh.flush()
+            subprocess.check_call(['bash', '-n', fh.name])
         sc = shellcode(["prog"], use_defaults=False, shell="tcsh", complete_arguments=["-o", "nospace"])
         sc = shellcode(["prog"], use_defaults=False, shell="woosh", complete_arguments=["-o", "nospace"])
         sc = shellcode(["prog"], shell="fish")
+        sc = shellcode(["prog"], shell="fish", argcomplete_script="~/.bash_completion.d/prog.py")
+
+    def test_option_help(self):
+        os.environ["_ARGCOMPLETE_DFS"] = "\t"
+        os.environ["_ARGCOMPLETE_SUPPRESS_SPACE"] = "1"
+        os.environ["_ARGCOMPLETE_SHELL"] = "fish"
+
+        p = ArgumentParser()
+        p.add_argument("--foo", help="foo" + IFS + "help")
+        p.add_argument("--bar", "--bar2", help="bar help")
+
+        subparsers = p.add_subparsers()
+        subparsers.add_parser("subcommand", help="subcommand help")
+        subparsers.add_parser("subcommand 2", help="subcommand 2 help")
+
+        completions = self.run_completer(p, "prog --f")
+        self.assertEqual(set(completions), {"--foo\tfoo help"})
+
+        completions = self.run_completer(p, "prog --b")
+        self.assertEqual(set(completions), {"--bar\tbar help", "--bar2\tbar help"})
+
+        completions = self.run_completer(p, "prog sub")
+        self.assertEqual(set(completions), {"subcommand\tsubcommand help", "subcommand 2\tsubcommand 2 help"})
+
+        os.environ["_ARGCOMPLETE_DFS"] = "invalid"
+        self.assertRaises(Exception, self.run_completer, p, "prog --b")
 
 class TestArgcompleteREPL(unittest.TestCase):
     def setUp(self):
@@ -890,8 +961,8 @@ class TestSplitLine(unittest.TestCase):
         self.assertEqual(self.prefix('a b c'), 'c')
 
     def test_escaped_special(self):
-        self.assertEqual(self.prefix('a\$b'), 'a$b')
-        self.assertEqual(self.prefix('a\`b'), 'a`b')
+        self.assertEqual(self.prefix(r'a\$b'), 'a$b')
+        self.assertEqual(self.prefix(r'a\`b'), 'a`b')
 
     def test_unescaped_special(self):
         self.assertEqual(self.prefix('a$b'), 'a$b')
@@ -899,8 +970,8 @@ class TestSplitLine(unittest.TestCase):
 
     @unittest.expectedFailure
     def test_escaped_special_in_double_quotes(self):
-        self.assertEqual(self.prefix('"a\$b'), 'a$b')
-        self.assertEqual(self.prefix('"a\`b'), 'a`b')
+        self.assertEqual(self.prefix(r'"a\$b'), 'a$b')
+        self.assertEqual(self.prefix(r'"a\`b'), 'a`b')
 
     def test_punctuation(self):
         self.assertEqual(self.prefix('a,'), 'a,')
@@ -1120,7 +1191,9 @@ class TestBash(_TestSh, unittest.TestCase):
         # This requires compopt which is not available in 3.x.
         expected_failures.append('test_quoted_exact')
 
-    install_cmd = 'eval "$(register-python-argcomplete prog)"'
+    # 'dummy' argument unused; checks multi-command registration works
+    # by passing 'prog' as the second argument.
+    install_cmd = 'eval "$(register-python-argcomplete dummy prog)"'
 
     def setUp(self):
         sh = pexpect.replwrap.bash()
@@ -1131,6 +1204,10 @@ class TestBash(_TestSh, unittest.TestCase):
         sh.run_command('complete -r python python2 python3')
         output = sh.run_command(self.install_cmd)
         self.assertEqual(output, '')
+        # Register a dummy completion with an external argcomplete script
+        # to ensure this doesn't overwrite our previous registration.
+        output = sh.run_command('eval "$(register-python-argcomplete dummy --external-argcomplete-script dummy)"')
+        self.assertEqual(output, '')
         self.sh = sh
 
     def test_one_space_after_exact(self):
@@ -1147,6 +1224,18 @@ class TestBash(_TestSh, unittest.TestCase):
         self.assertIn('PYTHON_ARGCOMPLETE_STDERR\r\n', output)
         self.assertTrue(output.endswith('foo\r\n'))
 
+    def test_temp_file(self):
+        self.sh.run_command('export ARGCOMPLETE_USE_TEMPFILES=1')
+        self.assertEqual(self.sh.run_command('prog basic f\t'), 'foo\r\n')
+        # Confirm we used a temp file by searching for the debug message.
+        self.sh.run_command('export _ARC_DEBUG=1')
+        output = self.sh.run_command('prog basic f\t')
+        self.assertIn('Using output file ', output)
+
+    def test_nounset(self):
+        self.sh.run_command('set -o nounset')
+        self.test_simple_completion()
+
 
 @unittest.skipIf(BASH_MAJOR_VERSION < 4, 'complete -D not supported')
 class TestBashGlobal(TestBash):
@@ -1235,7 +1324,13 @@ class TestTcsh(_TestSh, unittest.TestCase):
         path = ' '.join([os.path.join(BASE_DIR, 'scripts'), TEST_DIR, '$path'])
         sh.run_command('set path = ({0})'.format(path))
         sh.run_command('setenv PYTHONPATH {0}'.format(BASE_DIR))
-        output = sh.run_command('eval `register-python-argcomplete --shell tcsh prog`')
+        # 'dummy' argument unused; checks multi-command registration works
+        # by passing 'prog' as the second argument.
+        output = sh.run_command('eval `register-python-argcomplete --shell tcsh dummy prog`')
+        self.assertEqual(output, '')
+        # Register a dummy completion with an external argcomplete script
+        # to ensure this doesn't overwrite our previous registration.
+        output = sh.run_command('eval `register-python-argcomplete --shell tcsh dummy --external-argcomplete-script dummy`')
         self.assertEqual(output, '')
         self.sh = sh
 
@@ -1251,8 +1346,11 @@ class TestFish(_TestSh, unittest.TestCase):
     expected_failures = [
         'test_parse_special_characters',
         'test_comp_point',
-        'test_special_characters_double_quoted'
     ]
+    if FISH_VERSION_TUPLE < (3, 1):
+        expected_failures.extend([
+            'test_special_characters_double_quoted'
+        ])
 
     skipped = [
         'test_single_quotes_in_single_quotes',
@@ -1265,7 +1363,13 @@ class TestFish(_TestSh, unittest.TestCase):
         path = ' '.join([os.path.join(BASE_DIR, 'scripts'), TEST_DIR, '$PATH'])
         sh.run_command('set -x PATH {0}'.format(path))
         sh.run_command('set -x PYTHONPATH {0}'.format(BASE_DIR))
-        output = sh.run_command('register-python-argcomplete --shell fish prog | .')
+        # 'dummy' argument unused; checks multi-command registration works
+        # by passing 'prog' as the second argument.
+        output = sh.run_command('register-python-argcomplete --shell fish dummy prog | .')
+        self.assertEqual(output, '')
+        # Register a dummy completion with an external argcomplete script
+        # to ensure this doesn't overwrite our previous registration.
+        output = sh.run_command('register-python-argcomplete --shell fish dummy --external-argcomplete-script dummy | .')
         self.assertEqual(output, '')
         self.sh = sh
 
